#include <stdio.h>
#include <stdlib.h>

#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif

#include "config.h"

#define MAT_SIZE            1024
#define MAX_SOURCE_SIZE     65536


long check_results(T* A, T* B, T* C, int n) {
    long wrong = 0;
    int i, j, k;
    for (i=0; i<n; i++)
        for (j=0; j<n; j++) {
            T sum = 0;
            for (k=0; k<n; k++) sum += A[i*n+k]*B[k*n+j];
            if (sum != C[i*n+j]) wrong++;
        }
    return wrong;
}


cl_double get_duration(cl_event perf_event) {
	cl_ulong start = 0, end = 0;
	clGetEventProfilingInfo(perf_event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	clGetEventProfilingInfo(perf_event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	return (cl_double)(end - start)*(cl_double)(1e-09);
}


int main(int argc, char** argv) {
    
    int err;                            // error code returned from api calls
    int check = 0;                      // no correctness check by default
    unsigned int n = MAT_SIZE;          // matrix size
    
    T* A;                               // input matrix A on host
    T* B;                               // input matrix B on host
    T* C;                               // output matrix C on host
    long wrong;                         // number of wrong results returned
    char* program_name="matmul.cl";     // OpenCL C program name
    
	cl_platform_id platform_id;         // platform id
    cl_device_id device_id;             // compute device id 
    cl_context context;                 // compute context
    cl_command_queue commands;          // compute command queue
    cl_program program;                 // compute program
    cl_kernel kernel;                   // compute kernel
    cl_event perf_event;                // event object
    cl_double duration;                 // event duration
    
    cl_mem clA;                         // input matrix A on device
    cl_mem clB;                         // input matrix B on device
    cl_mem clC;                         // output matrix C on device
    
    // override defaults
    if (argc>1) n = atoi(argv[1]);
    if (argc>2) check = atoi(argv[2]);
    if (argc>3) program_name = argv[3];

    // round up matrix size to LOCAL_SIZE multiple
    n = ((n+LOCAL_SIZE-1)/LOCAL_SIZE)*LOCAL_SIZE;

    printf("Multiplying two %dx%d matrices\n", n, n);
    printf("Work group size: %d\n", LOCAL_SIZE);
    printf("Running OpenCL C program: %s\n", program_name);
    
    // allocate matrices on host
	A = malloc(n*n*sizeof(T));
	B = malloc(n*n*sizeof(T));
	C = malloc(n*n*sizeof(T));
   
    if (A == NULL || B == NULL || C == NULL) {
        printf("Error: Failed to allocate input/output buffers on host!\n");
        return EXIT_FAILURE;    
    }
    
    // Load the kernel source code into the array source_str
    FILE *fp;
    char *source_str;
    size_t source_size;
    
    fp = fopen(program_name, "r");
    if (!fp) {
        fprintf(stderr, "Error: Failed to load kernel!\n");
        exit(1);
    }
    source_str = (char*)malloc(MAX_SOURCE_SIZE);
    source_size = fread(source_str, 1, MAX_SOURCE_SIZE, fp);
    source_str[source_size] = 0;
    fclose( fp );
    
    // Fill our data set with random T values
    int i = 0;
    for(i = 0; i < n*n; i++) {
        A[i] = rand() % 10;
        B[i] = rand() % 10;
    }

    // Connect to a compute device
	err  = clGetPlatformIDs(1, &platform_id, NULL);
	err |= clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, 1, &device_id, NULL);
    
    if (err != CL_SUCCESS) {
        printf("Error: Failed to create a device group!\n");
        return EXIT_FAILURE;
    }
    
    // Create a compute context 
    context = clCreateContext(0, 1, &device_id, NULL, NULL, &err);
    if (!context) {
        printf("Error: Failed to create a compute context!\n");
        return EXIT_FAILURE;
    }
    
    // Create a command queue with profiling enabled
    commands = clCreateCommandQueue(context, device_id, CL_QUEUE_PROFILING_ENABLE, &err);
    if (!commands) {
        printf("Error: Failed to create a command commands!\n");
        return EXIT_FAILURE;
    }
    
    // Create the compute program from the source buffer
    program = clCreateProgramWithSource(context, 1, (const char **) & source_str, NULL, &err);
    if (!program) {
        printf("Error: Failed to create compute program!\n");
        return EXIT_FAILURE;
    }
    
    // Build the program executable
    err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
    if (err != CL_SUCCESS) {
        size_t len;
        char buffer[16384];
        
        printf("Error: Failed to build program executable!\n");
        clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
        fwrite(buffer, len, 1, stdout);
        exit(1);
    }
    
    // Create the compute kernel in the program we wish to run
    kernel = clCreateKernel(program, "matmul", &err);
    if (!kernel || err != CL_SUCCESS) {
        printf("Error: Failed to create compute kernel!\n");
        exit(1);
    }
    
    // Create the input and output arrays in device memory for our calculation
    clA = clCreateBuffer(context, CL_MEM_READ_ONLY,  sizeof(T)*n*n, NULL, NULL);
    clB = clCreateBuffer(context, CL_MEM_READ_ONLY,  sizeof(T)*n*n, NULL, NULL);
    clC = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(T)*n*n, NULL, NULL);
    if (!clA || !clB || !clC) {
        printf("Error: Failed to allocate device memory!\n");
        exit(1);
    }    
    
    // Write our data set into the input array in device memory 
    err = clEnqueueWriteBuffer(commands, clA, CL_TRUE, 0, sizeof(T)*n*n, A, 0, NULL, &perf_event);
    if (err != CL_SUCCESS) {
        printf("Error: Failed to write input matrix!\n");
        exit(1);
    }
    
	err = clFinish(commands);
    if (err) {
        printf("Error: Failed to complete data copy!\n");
        return EXIT_FAILURE;
    }
    
    duration = get_duration(perf_event);
    
    err = clEnqueueWriteBuffer(commands, clB, CL_TRUE, 0, sizeof(T)*n*n, B, 0, NULL, &perf_event);
    if (err != CL_SUCCESS) {
        printf("Error: Failed to write input matrix!\n");
        exit(1);
    }
    
	err = clFinish(commands);
    if (err) {
        printf("Error: Failed to complete data copy!\n");
        return EXIT_FAILURE;
    }
    
	duration += get_duration(perf_event);
	printf("Host -> device copy time: %.2f msec (%lu MB transferred, %.2f GB/sec)\n", duration*1000, 2*sizeof(T)*n*n/(1024*1024),
           2*sizeof(T)*n*n/(duration*1024*1024*1024));
    
    // Set the arguments to our compute kernel
    err = 0;
    err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &clA);
    err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &clB);
    err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), &clC);
    err |= clSetKernelArg(kernel, 3, sizeof(unsigned int), &n);
    if (err != CL_SUCCESS) {
        printf("Error: Failed to set kernel arguments! %d\n", err);
        exit(1);
    }
    
    // Execute the kernel over the entire range of our 2D input data set
    size_t global_dim[] = { n, n };
    size_t local_dim[]  = { LOCAL_SIZE, LOCAL_SIZE };
    err = clEnqueueNDRangeKernel(commands, kernel, 2, NULL, global_dim, local_dim, 0, NULL, &perf_event);
    if (err) {
        printf("Error: Failed to execute kernel!\n");
        return EXIT_FAILURE;
    }
    
    // Wait for kernel execution termination
	err = clFinish(commands);
    if (err) {
        printf("Error: Failed to complete kernel execution!\n");
        return EXIT_FAILURE;
    }
    
	duration = get_duration(perf_event);
	printf("Kernel execution time: %.2f msec (%.2f GFlops)\n", 
           duration*1000, 1e-09*n*n*n/duration);
    
    // Read back the results from the device to verify the output
    err = clEnqueueReadBuffer( commands, clC, CL_TRUE, 0, sizeof(T)*n*n, C, 0, NULL, &perf_event);  
    if (err != CL_SUCCESS) {
        printf("Error: Failed to read output array! %d\n", err);
        exit(1);
    }
    
	duration = get_duration(perf_event);
	printf("Device -> host copy time: %.2f msec (%lu MB transferred, %.2f GB/sec)\n", duration*1000, sizeof(T)*n*n/(1024*1024),
           sizeof(T)*n*n/(duration*1024*1024*1024));

    // Validate our results
    if (check) {
        printf("Performing correctness test... ");
        fflush(stdout);
        wrong = check_results(A, B, C, n);
        if (wrong) 
             printf("FAILED (%ld wrong items out of %ld)\n", wrong, (long)(n*n));
        else printf("PASSED (%ld items ok)\n", (long)(n*n));
    }
    
    // Shutdown and cleanup
    clReleaseMemObject(clA);
    clReleaseMemObject(clB);
    clReleaseMemObject(clC);
    clReleaseProgram(program);
    clReleaseKernel(kernel);
    clReleaseCommandQueue(commands);
    clReleaseContext(context);
    
    free(A);
    free(B);
    free(C);
    free(source_str);
    
    return 0;
}

