// =====================================================================
//  palind.c
// =====================================================================
 
//  Author:         (c) 2013 Camil Demetrescu
//  License:        See the end of this file for license information
//  Created:        May 26, 2013
 
//  Last changed:   $Date: 2013/05/26 13:01:07 $
//  Changed by:     $Author: demetres $
//  Revision:       $Revision: 1.00 $


#include <stdio.h>
#include <stdlib.h>

#include "clut.h"

#define ARRAY_SIZE      10000000
#define PROG_NAME       "palind.cl"
#define KERNEL_NAME     "palind"
#define LOCAL_SIZE      512


// ---------------------------------------------------------------------
// is_palindrome
// ---------------------------------------------------------------------
// sequential CPU version
int is_palindrome(int* v, int n, double* t) {
    int i, res = 1;
    double start = clut_get_real_time();
    for (i=0; i<n/2; i++) if (v[i] != v[n-i-1]) { res = 0; break; }
    *t = clut_get_real_time() - start;
    return res;
}


// ---------------------------------------------------------------------
// is_palindrome_cl
// ---------------------------------------------------------------------
// data-parallel GPU version
int is_palindrome_cl(int* v, int n, clut_device* dev, double* duration) {

    int       err;      // error code
    cl_kernel kernel;   // execution kernel
    cl_mem    dv;       // input array on device
    cl_mem    dres;     // output value on device
    int       res = 0;  // output value on host
    cl_event  evt;

    // create the compute kernel
    kernel = clCreateKernel(dev->program, KERNEL_NAME, &err);
    clut_check_err(err, "failed to create kernel");
    
    // allocate array on device as a copy of array on host
    dv = clCreateBuffer(dev->context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, n*sizeof(int), v, NULL);
    if (!dv) clut_panic("failed to allocate array on device memory");
    
    // allocate result buffer on device as a copy of res
    dres = clCreateBuffer(dev->context, CL_MEM_WRITE_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(int), &res, NULL);
    if (!dres) clut_panic("failed to allocate result buffer on device memory");

    // set the arguments to our compute kernel
    err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &dv);
    err  = clSetKernelArg(kernel, 1, sizeof(cl_mem), &dres);
    err |= clSetKernelArg(kernel, 2, sizeof(unsigned int), &n);
    clut_check_err(err, "failed to set kernel arguments");

    // execute the kernel over the half of the range of our 1D input data set
    size_t local_dim  = LOCAL_SIZE;
    size_t global_dim = n/2;
    global_dim = ((global_dim+LOCAL_SIZE-1)/LOCAL_SIZE)*LOCAL_SIZE; // round up

    err = clEnqueueNDRangeKernel(dev->queue, kernel, 1, 
                                 NULL, &global_dim, &local_dim, 0, NULL, &evt);
    clut_check_err(err, "failed to execute kernel");


    // copy result from device to host
    err = clEnqueueReadBuffer(dev->queue, dres, CL_TRUE, 0, 
                              sizeof(int), &res, 0, NULL, NULL);
    clut_check_err(err, "failed to read output result");

    // pass back operation duration
    if (duration != NULL) 
        *duration = clut_get_duration(evt);

    clReleaseMemObject(dv);
    clReleaseMemObject(dres);
    clReleaseKernel(kernel);

    return res == 0;
}


// ---------------------------------------------------------------------
// run_test
// ---------------------------------------------------------------------
void run_test(int* v, int n, clut_device* dev, char* msg) {

    double t, td;

    // check if the array is a palindrome on both host and device
    int res  = is_palindrome(v, n, &t);
    int resd = is_palindrome_cl(v, n, dev, &td);

    // validate our results
    printf("%s\n", msg);
    printf("    CPU took %.3f msec\n", t*1E03);
    printf("    Device took %.3f msec\n", td*1E03);
    printf("    Correctness test: %s\n", res == resd ? "PASSED" : "FAILED");    
}


// ---------------------------------------------------------------------
// main
// ---------------------------------------------------------------------
int main(int argc, char** argv) {

    clut_device dev;                // device structure
    int         n = ARRAY_SIZE;     // array size
    int*        v;                  // input array on host
    int         i;

    // override default array size
    if (argc > 1) n = atoi(argv[1]);

    // open device
    clut_open_device(&dev, PROG_NAME);

    printf("Palindrome test on array of size %d\n", n);
    printf("Work group size: %d\n", LOCAL_SIZE);
    clut_print_device_info(&dev);

    // allocate array on host
    v = malloc(n*sizeof(int));
    if (v == NULL) clut_panic("failed to allocate array on host");
    
    // make non-palindrome input array
    for (i = 0; i < n; i++) v[i] = 2*i;

    run_test(v, n, &dev,"Test 1 (non-palindrome):");

    // make palindrome input array
    for (i = 0; i < n/2; i++) v[i] = v[n-i-1] = 2*i;
  
    run_test(v, n, &dev,"Test 2 (palindrome):");

    // cleanup
    free(v);
    clut_close_device(&dev);
    
    return 0;
}


// Copyright (C) 2013 Camil Demetrescu
  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
  
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
  
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
// USA
