#include <stdio.h>
#include <stdlib.h>

#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif

#define MAX_SOURCE_SIZE 	0x10000         // 64 KB
#define GLOBAL_SIZE 		(16*1024*1024)  // 16*2^20 item per input array => 64MB of data
#define LOCAL_SIZE          8               // work items per work group


int main(void) {
    
    // create the two input vectors
    int i;
    int *A = (int*)malloc(sizeof(int)*GLOBAL_SIZE);
    int *B = (int*)malloc(sizeof(int)*GLOBAL_SIZE);
    for (i = 0; i < GLOBAL_SIZE; i++) {
        A[i] = i;
        B[i] = GLOBAL_SIZE - i;
    }
    
    // load the kernel source code into the array source_str
    FILE *fp = fopen("vecsum.cl", "r");
    if (!fp) exit((printf("Failed to load kernel.\n"),1));
    char* source_str = (char*)malloc(MAX_SOURCE_SIZE);
    size_t source_size = fread( source_str, 1, MAX_SOURCE_SIZE, fp);
    fclose( fp );
    
    // get platform and device information
    cl_platform_id platform_id = NULL;
    cl_device_id device_id = NULL;   
    cl_int ret = clGetPlatformIDs(1, &platform_id, NULL);
    if (ret) exit((printf("clGetPlatformIDs failed\n"),1));
    ret = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_DEFAULT, 1, &device_id, NULL);
    if (ret) exit((printf("clGetDeviceIDs failed\n"),1));

    // create an OpenCL context
    cl_context context = clCreateContext( NULL, 1, &device_id, NULL, NULL, &ret);
    if (ret) exit((printf("clCreateContext failed\n"),1));

    // create a command queue
    cl_command_queue command_queue = clCreateCommandQueue(context, device_id, 0, &ret);
    if (ret) exit((printf("clCreateCommandQueue failed\n"),1));

    // create memory buffers on the device for each vector 
    cl_mem a_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY, GLOBAL_SIZE * sizeof(int), NULL, &ret);
    if (ret) exit((printf("clCreateBuffer failed\n"),1));

    cl_mem b_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY, GLOBAL_SIZE * sizeof(int), NULL, &ret);
    if (ret) exit((printf("clCreateBuffer failed\n"),1));

    cl_mem c_mem_obj = clCreateBuffer(context, CL_MEM_WRITE_ONLY, GLOBAL_SIZE * sizeof(int), NULL, &ret);
    if (ret) exit((printf("clCreateBuffer failed\n"),1));

    // copy the lists A and B to their respective memory buffers
    ret  = clEnqueueWriteBuffer(command_queue, a_mem_obj, CL_TRUE, 0,
                               GLOBAL_SIZE * sizeof(int), A, 0, NULL, NULL);
    ret |= clEnqueueWriteBuffer(command_queue, b_mem_obj, CL_TRUE, 0, 
                               GLOBAL_SIZE * sizeof(int), B, 0, NULL, NULL);
    if (ret) exit((printf("clEnqueueWriteBuffer failed\n"),1));

    // create a program from the kernel source
    cl_program program = clCreateProgramWithSource(context, 1, 
                                                   (const char **)&source_str,
                                                   (const size_t *)&source_size, 
                                                   &ret);
    if (ret) exit((printf("clCreateProgramWithSource failed\n"),1));

    // build the program
    ret = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
    if (ret) {
        size_t len;
        char buffer[4096];        
        printf("clBuildProgram failed\n");
        clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
        printf("%s\n", buffer);
        exit(1);
    }

    // create an OpenCL kernel
    cl_kernel kernel = clCreateKernel(program, "vector_add", &ret);
    if (ret) exit((printf("clCreateKernel failed\n"),1));

    // set the arguments of the kernel
    ret  = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&a_mem_obj);
    ret |= clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&b_mem_obj);
    ret |= clSetKernelArg(kernel, 2, sizeof(cl_mem), (void *)&c_mem_obj);
    if (ret) exit((printf("clSetKernelArg failed\n"),1));

    // execute the OpenCL kernel on the list
    size_t global_item_size = GLOBAL_SIZE; // process the entire lists
    size_t local_item_size  = LOCAL_SIZE;  // divide work items into groups
    ret = clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL, 
                                 &global_item_size, &local_item_size, 0, NULL, NULL);
    if (ret) exit((printf("clEnqueueNDRangeKernel failed\n"),1));

    // read the memory buffer C on the device to the local variable C
    int* C = (int*)malloc(sizeof(int)*GLOBAL_SIZE);
    ret = clEnqueueReadBuffer(command_queue, c_mem_obj, CL_TRUE, 0, 
                              GLOBAL_SIZE * sizeof(int), C, 0, NULL, NULL);
    if (ret) exit((printf("clEnqueueReadBuffer failed\n"),1));

    // display the result to the screen
    for (i = 0; i < GLOBAL_SIZE; i++)
        if (A[i]+B[i]!=C[i]) break;
	if (i < GLOBAL_SIZE) printf("correctness test: FAILED\n");
	else                 printf("correctness test: PASSED\n");

    // clean up
    ret  = clReleaseKernel(kernel);
    ret |= clReleaseProgram(program);
    ret |= clReleaseMemObject(a_mem_obj);
    ret |= clReleaseMemObject(b_mem_obj);
    ret |= clReleaseMemObject(c_mem_obj);
    ret |= clReleaseCommandQueue(command_queue);
    ret |= clReleaseContext(context);
    if (ret) exit((printf("resource release failed\n"),1));

    free(A);
    free(B);
    free(C);
    return 0;
}
